package com.rudakova.media_archive.google;

import com.google.api.client.auth.oauth2.Credential;
import com.google.api.client.googleapis.auth.oauth2.GoogleAuthorizationCodeFlow;
import com.google.api.client.http.javanet.NetHttpTransport;
import com.google.api.client.json.JsonFactory;
import com.google.api.client.json.jackson2.JacksonFactory;
import com.google.api.services.drive.Drive;
import com.google.api.services.drive.model.About;
import com.google.api.services.drive.model.FileList;
import com.rudakova.media_archive.converter.FileConverter;
import com.rudakova.media_archive.dto.NotificationDto;
import com.rudakova.media_archive.model.Account;
import com.rudakova.media_archive.model.Channel;
import com.rudakova.media_archive.model.File;
import com.rudakova.media_archive.model.FolderList;
import com.rudakova.media_archive.repository.AccountRepository;
import com.rudakova.media_archive.repository.FileRepository;
import com.rudakova.media_archive.service.AuthService;
import com.rudakova.media_archive.service.RestService;
import com.rudakova.media_archive.util.FileUtil;
import com.rudakova.media_archive.util.HashCalculator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.core.env.Environment;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.stereotype.Service;
import org.springframework.web.client.RestTemplate;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.net.URL;
import java.time.Instant;
import java.util.*;
import java.util.stream.Collectors;

/**
 * Сервис для работы с Google Drive.
 */
@Service
public class GoogleService {
    private static final Logger LOG = LoggerFactory.getLogger(RestService.class);
    private static final String APPLICATION_NAME = "MediaArchive";
    private static final JsonFactory JSON_FACTORY = JacksonFactory.getDefaultInstance();
    private static final String FILE_FIELDS =
            "id, name, mimeType, parents, createdTime, trashed, size, webViewLink, hasThumbnail, thumbnailLink";

    @Autowired
    private Environment env;

    @Value("${filesFolder}")
    private String folder;

    @Value("${server.serverDomain}")
    private String serverDomain;

    @Autowired
    private NetHttpTransport HTTP_TRANSPORT;

    @Autowired
    private AccountRepository accountRepository;

    @Autowired
    private FileRepository fileRepository;

    @Autowired
    private AuthService authService;

    @Autowired
    private GoogleAuthorizationCodeFlow flow;

    @Autowired
    private RestTemplate restTemplate;

    private Credential getCredentials() throws IOException {
        String name = authService.getEmail();
        return flow.loadCredential(name == null ? "user" : name);
    }

    public AbstractMap<String, Object> getUserInfo() throws IOException {
        Drive service = getDrive();

        About about = service.about()
                .get()
                .setFields("user")
                .execute();

        return about.getUser();
    }

    public FolderList getFolders(String nextPageToken, String folderId) throws IOException {
        Drive service = getDrive();

        Account account = authService.getLoggedInAccount();
        folderId = folderId == null ? "root" : folderId;

        FileList result = service.files().list()
                .setQ("'" + folderId + "' in parents and mimeType = 'application/vnd.google-apps.folder'")
                .setPageToken(nextPageToken)
                .setPageSize(10)
                .setFields("nextPageToken, files(id, name, parents)")
                .execute();
        List<File> files = FileConverter.convertList(result, account, folderId);
        return new FolderList(result.getNextPageToken(), files);
    }

    public List<String> getFiles(String folderId) throws IOException {
        Drive service = getDrive();
        String nextPageToken = null;
        List<String> ids = new ArrayList<>();

        com.google.api.services.drive.model.File root = service.files()
                .get(folderId)
                .setFields(FILE_FIELDS)
                .execute();

        if (root.getTrashed()) {
            ids.add(root.getId());
        }

        do {
            FileList result = service.files().list()
                    .setQ("'" + folderId + "' in parents and trashed")
                    .setPageToken(nextPageToken)
                    .setFields("nextPageToken, files(id)")
                    .execute();
            ids.addAll(result.getFiles().stream()
                    .map(com.google.api.services.drive.model.File::getId)
                    .collect(Collectors.toList()));
            nextPageToken = result.getNextPageToken();
        } while (nextPageToken != null);

        return ids;
    }

    public void saveFilesFromDirectory() throws IOException {
        Drive service = getDrive();
        Account account = authService.getLoggedInAccount();
        String folderId, nextPageToken = null;
        Queue<String> folders = new ArrayDeque<>();

        com.google.api.services.drive.model.File root = service.files()
                .get(account.getDirectory())
                .setFields(FILE_FIELDS)
                .execute();
        saveFileToBd(null, account, folders, root);

        while (!folders.isEmpty()) {
            folderId = folders.poll();
            do {
                FileList result = service.files().list()
                        .setQ("'" + folderId + "' in parents and not trashed")
                        .setPageToken(nextPageToken)
                        .setPageSize(10)
                        .setFields("nextPageToken, files(" + FILE_FIELDS + ")")
                        .execute();
                nextPageToken = saveFilesToBd(folderId, account, folders, result);
            } while (nextPageToken != null);
        }
        accountRepository.save(account);
    }

    public String createNewFolder() throws IOException {
        Drive service = getDrive();

        com.google.api.services.drive.model.File fileMetadata = new com.google.api.services.drive.model.File();
        fileMetadata.setName("Медиаархив-выгрузка");
        fileMetadata.setMimeType("application/vnd.google-apps.folder");

        com.google.api.services.drive.model.File file = service.files().create(fileMetadata)
                .setFields("id")
                .execute();
        return file.getId();
    }

    public void moveToUploadDirectory(List<String> files) throws IOException {
        Drive service = getDrive();
        Account account = authService.getLoggedInAccount();
        String folderId = account.getUploadFolder();

        for (String id : files) {
            File fileFromBd = fileRepository.findById(id).get();
            fileFromBd.getParents().add(folderId);

            com.google.api.services.drive.model.File file =
                    FileConverter.convertToGoogleShortcut(fileFromBd, folderId);

            com.google.api.services.drive.model.File shortcutFile =
                    service.files().create(file)
                            .setFields("id")
                            .execute();

            fileFromBd.setShortcutId(shortcutFile.getId());
            fileRepository.save(fileFromBd);
        }
    }

    public void updateFile(String id) throws IOException {
        Drive service = getDrive();
        Account account = authService.getLoggedInAccount();

        com.google.api.services.drive.model.File file = service.files()
                .get(id)
                .setFields(FILE_FIELDS)
                .execute();
        saveFileToBd(file.getParents().get(0), account, null, file);
    }

    public int checkUpdate(String folderId, Instant date) throws IOException {
        Drive service = getDrive();
        Account account = authService.getLoggedInAccount();
        int count = 0;

        /*String modifiedTime = date.toString();
        modifiedTime = modifiedTime.substring(0, modifiedTime.lastIndexOf(":") + 3);
        modifiedTime = " and modifiedTime > '" + modifiedTime + "'";*/

        String nextPageToken = null, query;
        Queue<String> folders = new ArrayDeque<>();

        com.google.api.services.drive.model.File root = service.files()
                .get(folderId)
                .setFields(FILE_FIELDS)
                .execute();

        if (root.getTrashed()) {
            LOG.info("Root file: {}", root);
            saveFileToBd(root.getParents().get(0), account, folders, root);
            count++;
        } else {
            folders.offer(folderId);
        }

        while (!folders.isEmpty()) {
            folderId = folders.poll();
            query = "'" + folderId + "' in parents";

            do {
                FileList result = service.files().list()
                        .setQ(query)
                        .setPageToken(nextPageToken)
                        .setFields("nextPageToken, files(" + FILE_FIELDS + ")")
                        .execute();
                nextPageToken = saveFilesToBd(folderId, account, folders, result);
                count += result.getFiles().size();
            } while (nextPageToken != null);
        }

        return count;
    }

    public Drive getDrive() throws IOException {
        return new Drive.Builder(HTTP_TRANSPORT, JSON_FACTORY, getCredentials())
                .setApplicationName(APPLICATION_NAME)
                .build();
    }

    private String saveFilesToBd(String folderId, Account account, Queue<String> folders, FileList result) {
        String nextPageToken;
        for (com.google.api.services.drive.model.File file : result.getFiles()) {
            saveFileToBd(folderId, account, folders, file);
        }
        nextPageToken = result.getNextPageToken();
        return nextPageToken;
    }

    private void saveFileToBd(String folderId, Account account,
                                Queue<String> folders, com.google.api.services.drive.model.File file) {
        String mimeType = file.getMimeType();
        NotificationDto.Address address;
        File savedFile = fileRepository.findById(file.getId()).orElse(new File());
        if (mimeType.equals("application/vnd.google-apps.folder")) {
            folders.offer(file.getId());
            address = NotificationDto.Address.FOLDER;
            fileRepository.save(FileConverter.convertFromGoogleFile(file, savedFile, account, folderId));
        } else if (mimeType.contains("video")) {
            if (file.getHasThumbnail()) {
                savePreview(file);
            }
            address = NotificationDto.Address.VIDEO;
            fileRepository.save(FileConverter.convertFromGoogleFile(file, savedFile, account, folderId));
        } else {
            return;
        }

        if (Arrays.asList(env.getActiveProfiles()).contains("prod") && savedFile.getPreview() == null &&
                savedFile.isNeedChannel() && savedFile.getChannel() == null) {
            Channel channel = setWatchChannel(new NotificationDto(address), file.getId(), account);
            savedFile.setChannel(channel);
            fileRepository.save(savedFile);
        } else if (mimeType.contains("video") && savedFile.getChannel() != null && savedFile.getPreview() != null) {
            deleteWatchChannel(savedFile.getChannel(), account);
            savedFile.setChannel(null);
            savedFile.setNeedChannel(false);
            fileRepository.save(savedFile);
        }
    }

    private Channel setWatchChannel(NotificationDto notificationDto, String id, Account account) {
        final String url = "https://www.googleapis.com/drive/v3/files/{fileID}/watch";
        HttpHeaders httpHeaders = createHeader(account);

        HttpEntity<NotificationDto> entity = new HttpEntity<>(notificationDto, httpHeaders);
        LOG.info("Request: {}, {}, {}", id, entity.getBody(), entity.getHeaders());
        Channel channel = restTemplate.postForObject(url, entity, Channel.class, id);
        LOG.info(String.valueOf(channel));
        return channel;
    }

    private void deleteWatchChannel(Channel channel, Account account) {
        final String url = "https://www.googleapis.com/drive/v3/channels/stop";
        HttpHeaders httpHeaders = createHeader(account);

        HttpEntity<Channel> entity = new HttpEntity<>(channel, httpHeaders);
        LOG.info("Request: {}, {}", entity.getBody(), entity.getHeaders());
        String response = restTemplate.postForObject(url, entity, String.class);
        LOG.info(response);
    }

    public HttpHeaders createHeader(Account account) {
        HttpHeaders httpHeaders = new HttpHeaders();
        httpHeaders.setContentType(MediaType.APPLICATION_JSON);
        httpHeaders.setBearerAuth(account.getToken());
        return httpHeaders;
    }

    private void savePreview(com.google.api.services.drive.model.File file) {
        try {
            BufferedInputStream in =
                    new BufferedInputStream(new URL(file.getThumbnailLink()).openStream());
            byte[] data = in.readAllBytes();
            String hash = HashCalculator.getSHA256String(data);
            FileUtil.saveFile(hash, data, folder);
            file.setThumbnailLink(serverDomain + "/api/preview/" + hash);
        } catch (IOException e) {
            LOG.error("Error in save preview: {}", (Object) e.getStackTrace());
        }
    }
}
