package com.rudakova.media_archive.service;

import com.rudakova.media_archive.dto.AccountDto;
import com.rudakova.media_archive.dto.FileDto;
import com.rudakova.media_archive.dto.FileListDto;
import com.rudakova.media_archive.dto.FolderListDto;
import com.rudakova.media_archive.google.GoogleService;
import com.rudakova.media_archive.model.Account;
import com.rudakova.media_archive.model.File;
import com.rudakova.media_archive.model.FolderList;
import com.rudakova.media_archive.model.UpdateRecord;
import com.rudakova.media_archive.repository.AccountRepository;
import com.rudakova.media_archive.repository.FileRepository;
import com.rudakova.media_archive.repository.UpdateRecordRepository;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.*;

import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

/**
 * REST сервис.
 */
@RestController
@RequestMapping("/")
public class RestService {
    private static final Logger LOG = LoggerFactory.getLogger(RestService.class);
    private static final int PAGE_SIZE = 10;

    @Autowired
    private GoogleService googleService;

    @Autowired
    private FileRepository fileRepository;

    @Autowired
    private AuthService authService;

    @Autowired
    private AccountRepository accountRepository;

    @Autowired
    private UpdateRecordRepository updateRecordRepository;

    /**
     * Получение папок для выбора рабочей папки.
     * @param nextPageToken токен следующей страницы.
     * @param parent id родительской папки или ничего
     * @return список файлов.
     * @throws IOException
     */
    @GetMapping("/folders")
    public FolderListDto getFolders(@RequestParam(required= false) String nextPageToken,
                                    @RequestParam(required= false) String parent) throws IOException {
        FolderList folders = googleService.getFolders(nextPageToken, parent);
        return new FolderListDto(folders);
    }

    /**
     * Получение файлов из бд.
     * @param page номер страницы.
     * @param parent id родительской папки или ничего
     * @return список файлов.
     * @throws IOException
     */
    @GetMapping("/directory")
    public FileListDto getFiles(@RequestParam(required= false) Integer page,
                                @RequestParam String parent) {
        Account account = authService.getLoggedInAccount();
        page = page == null ? 0 : page;
        Pageable pageable = PageRequest.of(page, PAGE_SIZE);

        List<File> files = fileRepository.findByAccountAndParents(account.getEmailAddress(), parent, pageable);
        long count = fileRepository.countByAccountAndParent(account.getEmailAddress(), parent);
        int size = (int) ((count - 1) / PAGE_SIZE + 1);

        return new FileListDto(files, size);
    }

    /**
     * Сохранение рабочей папки.
     * @param folderId id выбранной папки в Google Drive.
     * @return id папки для выгрузки.
     * @throws IOException
     */
    @PostMapping("/directory")
    public String postWorkDirectory(@RequestParam String folderId) throws IOException {
        Account account = authService.getLoggedInAccount();
        account.setDirectory(folderId);
        account.setUploadFolder(googleService.createNewFolder());
        accountRepository.save(account);
        googleService.saveFilesFromDirectory();
        return account.getUploadFolder();
    }

    /**
     * Получение данных о пользователе.
     * @return пользовательские данные.
     */
    @GetMapping("/userInfo")
    public AccountDto getUserInfo() {
        Account account = authService.getLoggedInAccount();
        return new AccountDto(account);
    }

    /**
     * Применение изменений, которые пришли из Google Drive.
     */
    @PostMapping("/update")
    public void update() throws IOException {
        Account account = authService.getLoggedInAccount();
        List<UpdateRecord> records = updateRecordRepository.findByAccount(account.getEmailAddress());
        LOG.info("Records : {}", records.size());
        for (UpdateRecord record : records) {
            String type = record.getType();
            if (type.equals("trash")) {
                List<String> ids = googleService.getFiles(record.getFileId());
                for (String id : ids) {
                    fileRepository.deleteById(id);
                }
                LOG.info("Deleted : {}", ids.size());
            } else if (type.equals("properties")) {
                googleService.updateFile(record.getFileId());
                LOG.info("File was updated.");
            } else {
                int count = googleService.checkUpdate(record.getFileId(), record.getCreationDate());
                LOG.info("Added : {}", count);
            }
            updateRecordRepository.delete(record);
        }
    }

    /**
     * Перемещение выбранных файлов в папку для выгрузки.
     *
     * @param files id выбранных файлов.
     * @throws IOException
     */
    @PostMapping("/move")
    public void moveFilesToUploadDirectory(@RequestBody List<String> files) throws IOException {
        googleService.moveToUploadDirectory(files);
    }

    /**
     * Получение файлов по названию.
     *
     * @param name название файлов.
     */
    @PostMapping("/find")
    public FileListDto find(@RequestBody HashMap<String, String> tags,
                            @RequestParam(required = false) String name /*, @RequestParam int page*/) {
        //Pageable pageable = PageRequest.of(page, PAGE_SIZE);
        Account account = authService.getLoggedInAccount();
        String regexp = name == null ? ".*" : ".*" + name + ".*";
        List<File> files = fileRepository.findByName(account.getEmailAddress(), regexp);
        //long count = fileRepository.countByName(account.getEmailAddress(), regexp);
        List<File> filteredList = files.stream()
                .filter(file -> file.getTags().entrySet().containsAll(tags.entrySet()))
                .collect(Collectors.toList());
        int size = 1;

        return new FileListDto(filteredList, size);
    }

    /**
     * Добавление тегов к файлу.
     *
     * @param tags список новых тегов.
     * @param id   файла.
     * @return модель файла.
     */
    @PostMapping("/add")
    public ResponseEntity<FileDto> addTags(@RequestBody HashMap<String, Object> tags, @RequestParam String id) {
        Optional<File> fileOptional = fileRepository.findById(id);
        if (fileOptional.isEmpty()) {
            return new ResponseEntity<>(HttpStatus.BAD_REQUEST);
        }
        File file = fileOptional.get();
        file.addTags(tags);
        fileRepository.save(file);
        return new ResponseEntity<>(new FileDto(file), HttpStatus.OK);
    }
}
